@@ -2,7 +2,7 @@
  *
  * This file is part of PRoot.
  *
- * Copyright (C) 2014 STMicroelectronics
+ * Copyright (C) 2015 STMicroelectronics
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,6 +31,7 @@
 #include <signal.h>     /* kill(2), SIGKILL, */
 #include <sys/ptrace.h> /* ptrace(2), PTRACE_*, */
 #include <errno.h>      /* E*, */
+#include <inttypes.h>   /* PRI*, */
 
 #include "tracee/tracee.h"
 #include "tracee/reg.h"
@@ -44,6 +45,10 @@
 
 #include "compat.h"
 
+#ifndef __W_STOPCODE
+#define __W_STOPCODE(sig)	((sig) <<8 | 0x7f)
+#endif
+
 typedef LIST_HEAD(tracees, tracee) Tracees;
 static Tracees tracees;
 
@@ -203,6 +208,8 @@ no_mem:
 	return NULL;
 }
 
+static uint64_t next_vpid = 1;
+
 /**
  * Allocate a new entry for the tracee @pid, then set its destructor
  * and add it to the list of tracees.  This function returns NULL if
@@ -220,6 +227,8 @@ static Tracee *new_tracee(pid_t pid)
 	talloc_set_destructor(tracee, remove_tracee);
 
 	tracee->pid = pid;
+	tracee->vpid = next_vpid++;
+
 	LIST_INSERT_HEAD(&tracees, tracee, link);
 
 	tracee->life_context = talloc_new(tracee);
@@ -234,7 +243,7 @@ static Tracee *new_tracee(pid_t pid)
  * wait(2) manual for the meaning of @wait_options.  This function
  * returns NULL if there's no such ptracee.
  */
-static Tracee *get_ptracee(const Tracee *ptracer, pid_t pid, bool only_stopped,
+Tracee *get_ptracee(const Tracee *ptracer, pid_t pid, bool only_stopped,
 			bool only_with_pevent, word_t wait_options)
 {
 	Tracee *ptracee;
@@ -335,6 +344,22 @@ Tracee *get_tracee(const Tracee *current_tracee, pid_t pid, bool create)
 	return (create ? new_tracee(pid) : NULL);
 }
 
+/**
+ * Mark tracee as terminated and optionally take action.
+ */
+void terminate_tracee(Tracee *tracee)
+{
+        tracee->terminated = true;
+
+        /* Case where the terminated tracee is marked
+           to kill all tracees on exit.
+        */
+        if (tracee->killall_on_exit) {
+                VERBOSE(tracee, 1, "terminating all tracees on exit");
+                kill_all_tracees();
+        }
+}
+
 /**
  * Free all tracees marked as terminated.
  */
@@ -551,6 +576,8 @@ int new_child(Tracee *parent, word_t clone_flags)
 			(void) restart_tracee(child, 0);
 	}
 
+	VERBOSE(child, 1, "vpid %" PRIu64 ": pid %d", child->vpid, child->pid);
+
 	return 0;
 }
 
